"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.myQGarageDoor = void 0;
/* Copyright(C) 2017-2022, HJD (https://github.com/hjdhjd). All rights reserved.
 *
 * myq-garagedoor.ts: Garage door device class for myQ.
 */
const settings_1 = require("./settings");
const myq_accessory_1 = require("./myq-accessory");
class myQGarageDoor extends myq_accessory_1.myQAccessory {
    // Configure a garage door accessory for HomeKit.
    configureDevice() {
        // Initialize.
        this.batteryDeviceSupport = false;
        this.obstructionDetected = false;
        // Save our context information before we wipe it out.
        const device = this.accessory.context.device;
        const doorInitialState = this.accessory.context.doorState;
        // Clean out the context object.
        this.accessory.context = {};
        this.accessory.context.device = device;
        this.accessory.context.doorState = doorInitialState;
        this.configureInfo();
        this.configureGarageDoor();
        this.configureBatteryInfo();
        this.configureMqtt();
    }
    // Configure the garage door service for HomeKit.
    configureGarageDoor() {
        var _a;
        const device = this.accessory.context.device;
        let garagedoorService = this.accessory.getService(this.hap.Service.GarageDoorOpener);
        // Add the garage door opener service to the accessory, if needed.
        if (!garagedoorService) {
            garagedoorService = new this.hap.Service.GarageDoorOpener((_a = this.accessory.displayName) !== null && _a !== void 0 ? _a : device.name);
            this.accessory.addService(garagedoorService);
        }
        // The initial door state when we first startup. The bias functions will help us
        // figure out what to do if we're caught in a tweener state.
        const doorCurrentState = this.doorCurrentStateBias(this.accessory.context.doorState);
        const doorTargetState = this.doorTargetStateBias(doorCurrentState);
        // Add all the events to our accessory so we can act on HomeKit actions. We also set the current and target door states
        // based on our saved state from previous sessions.
        garagedoorService
            .updateCharacteristic(this.hap.Characteristic.CurrentDoorState, doorCurrentState)
            .updateCharacteristic(this.hap.Characteristic.TargetDoorState, doorTargetState)
            .getCharacteristic(this.hap.Characteristic.TargetDoorState)
            .onSet((value) => {
            this.setDoorState(value);
        });
        // Add all the events to our accessory so we can tell HomeKit our state.
        garagedoorService
            .getCharacteristic(this.hap.Characteristic.CurrentDoorState)
            .onGet(() => {
            if (this.doorStatus() === -1) {
                new Error("Unable to determine the current door state.");
            }
            // Return garage door status.
            return this.doorStatus();
        });
        // Make sure we can detect obstructions.
        garagedoorService
            .getCharacteristic(this.hap.Characteristic.ObstructionDetected)
            .onGet(() => {
            // For a refresh of the door status, but we're really unconcerned about what it returns here.
            this.doorStatus();
            // See if we have an obstruction to alert on.
            if (this.obstructionDetected) {
                this.log.info("%s: Obstruction detected.", this.accessory.displayName);
            }
            return this.obstructionDetected;
        });
        return true;
    }
    // Configure the battery status information for HomeKit.
    configureBatteryInfo() {
        // If we don't have a door position sensor, we're done.
        if (this.doorPositionSensorBatteryStatus() === -1) {
            return false;
        }
        const gdService = this.accessory.getService(this.hap.Service.GarageDoorOpener);
        // Verify we've already setup the garage door service before trying to configure it.
        if (!gdService) {
            return false;
        }
        // Add the battery service if we don't have it, so we can report on battery state.
        const gdBatteryService = this.accessory.getService(this.hap.Service.Battery) ||
            this.accessory.addService(this.hap.Service.Battery);
        if (!gdBatteryService) {
            this.log.error("%s: Unable to add battery status support.", this.accessory.displayName);
            return false;
        }
        gdBatteryService
            .getCharacteristic(this.hap.Characteristic.StatusLowBattery)
            .onGet(() => {
            return this.doorPositionSensorBatteryStatus();
        });
        // We only want to configure this once, not on each update.
        // Not the most elegant solution, but it gets the job done.
        this.batteryDeviceSupport = true;
        this.log.info("%s: Door position sensor detected. Enabling battery status support.", this.accessory.displayName);
        return true;
    }
    // Configure MQTT.
    configureMqtt() {
        var _a, _b;
        // Return the current status of the garage door.
        (_a = this.platform.mqtt) === null || _a === void 0 ? void 0 : _a.subscribe(this.accessory, "garagedoor/get", (message) => {
            var _a, _b;
            const value = (_a = message === null || message === void 0 ? void 0 : message.toString()) === null || _a === void 0 ? void 0 : _a.toLowerCase();
            // When we get the right message, we return the list of liveviews.
            if (value !== "true") {
                return;
            }
            // Publish the state of the garage door.
            (_b = this.platform.mqtt) === null || _b === void 0 ? void 0 : _b.publish(this.accessory, "garagedoor", this.translateDoorState(this.doorStatus()));
            this.log.info("%s: Garage door status published via MQTT.", this.accessory.displayName);
        });
        // Return the current status of the garage door.
        (_b = this.platform.mqtt) === null || _b === void 0 ? void 0 : _b.subscribe(this.accessory, "garagedoor/set", (message) => {
            var _a;
            const value = (_a = message === null || message === void 0 ? void 0 : message.toString()) === null || _a === void 0 ? void 0 : _a.toLowerCase();
            let targetName;
            let targetState;
            // Figure out what we're setting to.
            switch (value) {
                case "open":
                    targetState = this.hap.Characteristic.TargetDoorState.OPEN;
                    targetName = "Open";
                    break;
                case "close":
                    targetState = this.hap.Characteristic.TargetDoorState.CLOSED;
                    targetName = "Close";
                    break;
                default:
                    this.log.error("%s: Unknown door command received via MQTT: %s.", this.accessory.displayName, value);
                    return;
            }
            // Move the door to the desired position.
            if (this.setDoorState(targetState)) {
                this.log.info("%s: %s command received via MQTT.", this.accessory.displayName, targetName);
                return;
            }
            this.log.error("%s: Error executing door command via MQTT: %s.", this.accessory.displayName, value);
        });
    }
    // Open or close the garage door.
    setDoorState(value) {
        var _a, _b;
        const myQState = this.doorStatus();
        const accessory = this.accessory;
        const hap = this.hap;
        if (myQState === -1) {
            //new Error("Unable to determine the current door state."));
            return false;
        }
        // If we are already opening or closing the garage door, we error out. myQ doesn't appear to allow
        // interruptions to an open or close command that is currently executing - it must be allowed to
        // complete its action before accepting a new one.
        if ((myQState === hap.Characteristic.CurrentDoorState.OPENING) || (myQState === hap.Characteristic.CurrentDoorState.CLOSING)) {
            const actionExisting = myQState === hap.Characteristic.CurrentDoorState.OPENING ? "opening" : "closing";
            const actionAttempt = value === hap.Characteristic.TargetDoorState.CLOSED ? "close" : "open";
            this.log.error("%s: Unable to %s door while currently attempting to complete %s. myQ must complete it's existing action before attempting a new one.", accessory.displayName, actionAttempt, actionExisting);
            return false;
        }
        // Close the garage door.
        if (value === hap.Characteristic.TargetDoorState.CLOSED) {
            // HomeKit is asking us to close the garage door, but let's make sure it's not already closed first.
            if (myQState !== hap.Characteristic.CurrentDoorState.CLOSED) {
                // We set this to closing instead of closed because we want to show state transitions in HomeKit. In
                // addition, myQ won't immediately execute this command for safety reasons - it enforces a warning tone
                // for a few seconds before it starts the action.
                (_a = accessory
                    .getService(hap.Service.GarageDoorOpener)) === null || _a === void 0 ? void 0 : _a.updateCharacteristic(hap.Characteristic.CurrentDoorState, hap.Characteristic.CurrentDoorState.CLOSING);
                // Execute this command and begin polling myQ for state changes.
                void this.doorCommand(hap.Characteristic.TargetDoorState.CLOSED);
            }
            return true;
        }
        // Open the garage door.
        if (value === hap.Characteristic.TargetDoorState.OPEN) {
            // HomeKit is informing us to open the door, but we don't want to act if it's already open.
            if (myQState !== hap.Characteristic.CurrentDoorState.OPEN) {
                // We set this to opening instad of open because we want to show our state transitions to HomeKit and end users.
                (_b = accessory
                    .getService(hap.Service.GarageDoorOpener)) === null || _b === void 0 ? void 0 : _b.updateCharacteristic(hap.Characteristic.CurrentDoorState, hap.Characteristic.CurrentDoorState.OPENING);
                // Execute this command and begin polling myQ for state changes.
                void this.doorCommand(hap.Characteristic.TargetDoorState.OPEN);
            }
            return true;
        }
        // HomeKit has told us something that we don't know how to handle.
        this.log.error("%s: Unknown SET event received: %s.", accessory.displayName, value);
        return false;
    }
    // Update our HomeKit status.
    updateState() {
        var _a, _b, _c, _d;
        const accessory = this.accessory;
        const hap = this.hap;
        const oldState = accessory.context.doorState;
        const myQState = this.doorStatus();
        // If we can't get our status, we're probably not able to connect to the myQ API.
        if (myQState === -1) {
            this.log.error("%s: Unable to determine the current door state.", accessory.displayName);
            return false;
        }
        // Update the state in HomeKit
        if (oldState !== myQState) {
            accessory.context.doorState = myQState;
            // We are only going to update the target state if our current state is NOT stopped. If we are stopped,
            // we are at the target state by definition. Unfortunately, the iOS Home app doesn't seem to correctly
            // report a stopped state, although you can find it correctly reported in other HomeKit apps like Eve Home.
            // Finally, we want to ensure we update TargetDoorState before updating CurrentDoorState in order to work
            // around some notification quirks HomeKit occasionally has.
            if (myQState !== hap.Characteristic.CurrentDoorState.STOPPED) {
                const targetState = this.doorTargetStateBias(myQState);
                (_a = accessory.getService(hap.Service.GarageDoorOpener)) === null || _a === void 0 ? void 0 : _a.updateCharacteristic(hap.Characteristic.TargetDoorState, targetState);
            }
            (_b = accessory.getService(hap.Service.GarageDoorOpener)) === null || _b === void 0 ? void 0 : _b.updateCharacteristic(hap.Characteristic.CurrentDoorState, myQState);
            // When we detect any state change, we want to increase our polling resolution to provide timely updates.
            this.platform.pollOptions.count = 0;
            this.platform.poll(this.config.refreshInterval * -1);
            this.log.info("%s: %s.", accessory.displayName, this.translateDoorState(myQState));
            // Publish to MQTT, if the user has configured it.
            (_c = this.platform.mqtt) === null || _c === void 0 ? void 0 : _c.publish(accessory, "garagedoor", this.translateDoorState(myQState).toLowerCase());
        }
        // Update battery status only if it's supported by the device.
        if (this.batteryDeviceSupport) {
            const batteryStatus = this.doorPositionSensorBatteryStatus();
            // Update our battery state.
            if (batteryStatus !== -1) {
                (_d = accessory.getService(hap.Service.Battery)) === null || _d === void 0 ? void 0 : _d.updateCharacteristic(hap.Characteristic.StatusLowBattery, batteryStatus);
            }
        }
        return true;
    }
    // Return the status of the door. This function maps myQ door status to HomeKit door status.
    doorStatus() {
        // Door state cheat sheet.
        // autoreverse is how the myQ API communicated an obstruction...go figure. Unfortunately, it
        // only seems to last the duration of the door reopening (reversal).
        const doorStates = {
            autoreverse: settings_1.MYQ_OBSTRUCTED,
            closed: this.hap.Characteristic.CurrentDoorState.CLOSED,
            closing: this.hap.Characteristic.CurrentDoorState.CLOSING,
            open: this.hap.Characteristic.CurrentDoorState.OPEN,
            opening: this.hap.Characteristic.CurrentDoorState.OPENING,
            stopped: this.hap.Characteristic.CurrentDoorState.STOPPED
        };
        const device = this.accessory.context.device;
        if (!device) {
            this.log.error("%s: Can't find the associated device in the myQ API.", this.accessory.displayName);
            return -1;
        }
        // Retrieve the door state from myQ and map it to HomeKit.
        const myQState = doorStates[device.state.door_state];
        if (myQState === undefined) {
            this.log.error("%s: Unknown door state encountered: %s.", this.accessory.displayName, device.state.door_state);
            return -1;
        }
        // Obstructed states in the myQ API remain active for a very small period of time. Furthermore, the way
        // HomeKit informs you of an obstructed state is through a status update on the Home app home screen.
        // This ultimately means that an obstructed state has a very small chance of actually being visible to
        // a user unless they happen to be looking at the Home app at the exact moment the obstruction is detected.
        // To ensure the user has a reasonable chance to notice the obstructed state, we will alert a user for up
        // to MYQ_OBSTRUCTION_ALERT_DURATION seconds after the last time we detected an obstruction before clearing
        // out the alert.
        if (myQState === settings_1.MYQ_OBSTRUCTED) {
            // Clear any other timer that might be out there for obstructions.
            clearTimeout(this.ObstructionTimer);
            // Obstruction detected.
            this.obstructionDetected = true;
            const accessory = this.accessory;
            const hap = this.hap;
            // Set the timer for clearing out the obstruction state.
            this.ObstructionTimer = setTimeout(() => {
                var _a;
                this.obstructionDetected = false;
                (_a = accessory
                    .getService(hap.Service.GarageDoorOpener)) === null || _a === void 0 ? void 0 : _a.updateCharacteristic(hap.Characteristic.ObstructionDetected, this.obstructionDetected);
                this.log.info("%s: Obstruction cleared.", this.accessory.displayName);
            }, settings_1.MYQ_OBSTRUCTION_ALERT_DURATION * 1000);
        }
        return myQState;
    }
    // Execute garage door commands.
    async doorCommand(command) {
        let myQCommand;
        // Translate the command from HomeKit to myQ.
        switch (command) {
            case this.hap.Characteristic.TargetDoorState.OPEN:
                myQCommand = "open";
                break;
            case this.hap.Characteristic.TargetDoorState.CLOSED:
                myQCommand = "close";
                break;
            default:
                this.log.error("%s: Unknown door command encountered: %s.", this.accessory.displayName, command);
                return false;
        }
        return super.command(myQCommand);
    }
    // Decode HomeKit door state in user-friendly terms.
    translateDoorState(state) {
        // HomeKit state decoder ring.
        switch (state) {
            case this.hap.Characteristic.CurrentDoorState.OPEN:
                return "Open";
            case this.hap.Characteristic.CurrentDoorState.CLOSED:
                return "Closed";
            case this.hap.Characteristic.CurrentDoorState.OPENING:
                return "Opening";
            case this.hap.Characteristic.CurrentDoorState.CLOSING:
                return "Closing";
            case this.hap.Characteristic.CurrentDoorState.STOPPED:
                return "Stopped";
            case settings_1.MYQ_OBSTRUCTED:
                return "Obstructed";
            default:
                return "Unknown";
        }
    }
    // Return our bias for what the current door state should be. This is primarily used for our initial bias on startup.
    doorCurrentStateBias(myQState) {
        // Our current state reflects having to take an opinion on what open or closed means to
        // HomeKit. For the obvious states, this is easy. For some of the edge cases, it can be less so.
        // Our north star is that if we are in an obstructed state, we are open.
        switch (myQState) {
            case this.hap.Characteristic.CurrentDoorState.OPEN:
            case this.hap.Characteristic.CurrentDoorState.OPENING:
            case settings_1.MYQ_OBSTRUCTED:
                return this.hap.Characteristic.CurrentDoorState.OPEN;
            case this.hap.Characteristic.CurrentDoorState.STOPPED:
                return this.hap.Characteristic.CurrentDoorState.STOPPED;
            case this.hap.Characteristic.CurrentDoorState.CLOSED:
            case this.hap.Characteristic.CurrentDoorState.CLOSING:
            default:
                return this.hap.Characteristic.CurrentDoorState.CLOSED;
        }
    }
    // Return our bias for what the target door state should be.
    doorTargetStateBias(myQState) {
        // We need to be careful with respect to the target state and we need to make some
        // reasonable assumptions about where we intend to end up. If we are opening or closing,
        // our target state needs to be the completion of those actions. If we're stopped or
        // obstructed, we're going to assume the desired target state is to be open, since that
        // is the typical garage door behavior, and it's impossible for us to know with reasonable
        // certainty what the original intention of the action was.
        switch (myQState) {
            case this.hap.Characteristic.CurrentDoorState.OPEN:
            case this.hap.Characteristic.CurrentDoorState.OPENING:
            case this.hap.Characteristic.CurrentDoorState.STOPPED:
            case settings_1.MYQ_OBSTRUCTED:
                return this.hap.Characteristic.TargetDoorState.OPEN;
            case this.hap.Characteristic.CurrentDoorState.CLOSED:
            case this.hap.Characteristic.CurrentDoorState.CLOSING:
            default:
                return this.hap.Characteristic.TargetDoorState.CLOSED;
        }
    }
    // Return the battery status of the door sensor, if supported on the device.
    doorPositionSensorBatteryStatus() {
        const device = this.accessory.context.device;
        if (!device) {
            this.log.error("%s: Can't find the associated device in the myQ API.", this.accessory.displayName, this.accessory.UUID);
            return -1;
        }
        // If we don't find the dps_low_battery_mode attribute, then this device may not support it.
        if (!("state" in device) || !("dps_low_battery_mode" in device.state)) {
            return -1;
        }
        return device.state.dps_low_battery_mode ? this.hap.Characteristic.StatusLowBattery.BATTERY_LEVEL_LOW :
            this.hap.Characteristic.StatusLowBattery.BATTERY_LEVEL_NORMAL;
    }
}
exports.myQGarageDoor = myQGarageDoor;
//# sourceMappingURL=myq-garagedoor.js.map